"""Tests for :mod:`.integration.api`."""

from unittest import TestCase, mock
from . import service, status, exceptions

from flask import Flask


class TestHTTPIntegration(TestCase):
    """Test for :class:`.HTTPIntegration`."""

    def session(self, status_code=status.OK, method="get", json={},
                content="", headers={}):
        """Make a mock session."""
        return mock.MagicMock(**{
            method: mock.MagicMock(
                return_value=mock.MagicMock(
                    status_code=status_code,
                    json=mock.MagicMock(
                        return_value=json
                    ),
                    content=content,
                    headers=headers
                )
            )
        })

    def setUp(self):
        """Make some apps."""
        self.app_one = Flask('one')
        self.app_two = Flask('two')

    def test_init_app(self):
        """:func:`HTTPIntegration.init_app` sets config defaults."""
        service.HTTPIntegration.init_app(self.app_one)
        self.assertEqual(self.app_one.config['BASE_ENDPOINT'],
                         service.DEFAULT_ENDPOINT)
        self.assertEqual(self.app_one.config['BASE_VERIFY'],
                         service.DEFAULT_VERIFY)

    @mock.patch(f'{service.__name__}.requests.Session')
    def test_request_in_context(self, mock_Session):
        """:func:`HTTPIntegration.request` uses contextual config."""
        session_one = mock.MagicMock(
            get=mock.MagicMock(
                return_value=mock.MagicMock(status_code=status.OK)
            )
        )
        session_two = mock.MagicMock(
            get=mock.MagicMock(
                return_value=mock.MagicMock(status_code=status.OK)
            )
        )
        mock_Session.side_effect = [session_one, session_two]
        self.app_one.config['BASE_ENDPOINT'] = 'https://fooservice/baz'
        service.HTTPIntegration.init_app(self.app_one)
        self.app_two.config['BASE_ENDPOINT'] = 'https://bazservice/foo'
        service.HTTPIntegration.init_app(self.app_two)

        with self.app_one.app_context():
            service.HTTPIntegration.request('get', '/one')

        with self.app_two.app_context():
            service.HTTPIntegration.request('get', '/two')

        session_one.get.assert_called_with('https://fooservice/baz/one')
        self.assertEqual(session_one.get.call_count, 1)
        session_two.get.assert_called_with('https://bazservice/foo/two')
        self.assertEqual(session_two.get.call_count, 1)

    @mock.patch(f'{service.__name__}.requests.Session')
    def test_handles_500_series(self, mock_Session):
        """Handles 500-series response codes."""
        session_one = self.session(status.SERVICE_UNAVAILABLE)
        mock_Session.return_value = session_one
        self.app_one.config['BASE_ENDPOINT'] = 'https://fooservice/baz'
        service.HTTPIntegration.init_app(self.app_one)

        with self.app_one.app_context():
            try:
                service.HTTPIntegration.request('get', '/one')
            except Exception as ex:
                self.assertIsInstance(ex, exceptions.RequestFailed)
                self.assertEqual(ex.status_code,
                                 status.SERVICE_UNAVAILABLE)

    @mock.patch(f'{service.__name__}.requests.Session')
    def test_handles_unauthorized(self, mock_Session):
        """Handles 401 Unauthorized response code."""
        session_one = self.session(status.UNAUTHORIZED)
        mock_Session.return_value = session_one
        self.app_one.config['BASE_ENDPOINT'] = 'https://fooservice/baz'
        service.HTTPIntegration.init_app(self.app_one)

        with self.app_one.app_context():
            try:
                service.HTTPIntegration.request('get', '/one')
            except Exception as ex:
                self.assertIsInstance(ex, exceptions.RequestUnauthorized)
                self.assertEqual(ex.status_code,
                                 status.UNAUTHORIZED)

    @mock.patch(f'{service.__name__}.requests.Session')
    def test_handles_forbidden(self, mock_Session):
        """Handles 403 Forbidden response code."""
        session_one = self.session(status.FORBIDDEN)
        mock_Session.return_value = session_one
        self.app_one.config['BASE_ENDPOINT'] = 'https://fooservice/baz'
        service.HTTPIntegration.init_app(self.app_one)

        with self.app_one.app_context():
            try:
                service.HTTPIntegration.request('get', '/one')
            except Exception as ex:
                self.assertIsInstance(ex, exceptions.RequestForbidden)
                self.assertEqual(ex.status_code,
                                 status.FORBIDDEN)

    @mock.patch(f'{service.__name__}.requests.Session')
    def test_handles_not_found(self, mock_Session):
        """Handles 404 Not Found response code."""
        session_one = self.session(status.NOT_FOUND)
        mock_Session.return_value = session_one
        self.app_one.config['BASE_ENDPOINT'] = 'https://fooservice/baz'
        service.HTTPIntegration.init_app(self.app_one)

        with self.app_one.app_context():
            try:
                service.HTTPIntegration.request('get', '/one')
            except Exception as ex:
                self.assertIsInstance(ex, exceptions.NotFound)
                self.assertEqual(ex.status_code,
                                 status.NOT_FOUND)

    @mock.patch(f'{service.__name__}.requests.Session')
    def test_handle_bad_request(self, mock_Session):
        """Handles 400 Bad Request response code."""
        session_one = self.session(status.BAD_REQUEST)
        mock_Session.return_value = session_one
        self.app_one.config['BASE_ENDPOINT'] = 'https://fooservice/baz'
        service.HTTPIntegration.init_app(self.app_one)

        with self.app_one.app_context():
            try:
                service.HTTPIntegration.request('get', '/one')
            except Exception as ex:
                self.assertIsInstance(ex, exceptions.BadRequest)
                self.assertEqual(ex.status_code,
                                 status.BAD_REQUEST)

    @mock.patch(f'{service.__name__}.requests.Session')
    def test_handle_unexpected(self, mock_Session):
        """Handles unexpected response code."""
        session_one = self.session(status.USE_PROXY)
        mock_Session.return_value = session_one
        self.app_one.config['BASE_ENDPOINT'] = 'https://fooservice/baz'
        service.HTTPIntegration.init_app(self.app_one)

        with self.app_one.app_context():
            try:
                service.HTTPIntegration.request('get', '/one')
            except Exception as ex:
                self.assertIsInstance(ex, exceptions.RequestFailed)
                self.assertEqual(ex.status_code,
                                 status.USE_PROXY)

        session_one = self.session(status.OK)
        mock_Session.return_value = session_one
        self.app_one.config['BASE_ENDPOINT'] = 'https://fooservice/baz'
        service.HTTPIntegration.init_app(self.app_one)

        with self.app_one.app_context():
            try:
                service.HTTPIntegration.request(
                    'get',
                    '/one',
                    expected_code=[status.GONE]
                )
            except Exception as ex:
                self.assertIsInstance(ex, exceptions.RequestFailed)
                self.assertEqual(ex.status_code, status.OK)



"""
Helpers for working with Flask globals.

Flask makes heavy use of proxy objects. Depending on the context in which code
is being executed, many of the Flask globals will refer to different things
(or sometimes nothing at all). It is nice to be able to write code without
handling the various scenarios that this creates. That's what this module is
for: instead of accessing proxies directly, the functions here will access
proxies appropriately depending on the context, and return the appropriate
object.
"""

import os
from typing import Optional, Union, Mapping
from flask import g, Flask
from flask import current_app as flask_app
import werkzeug


def get_application_config(app: Optional[Flask] = None) ->Mapping:
    """
    Get a configuration from the current app, or fall back to os.env.

    Parameters
    ----------
    app : :class:`flask.Flask`

    Returns
    -------
    dict-like
        This is either the current Flask application configuration, or
        ``os.environ``. Either of these should support the ``get()`` method.
    """
    # pylint: disable=protected-access
    if app is not None:
        if isinstance(app, Flask):
            return app.config
    if flask_app:    # Proxy object; falsey if there is no application context.
        return flask_app.config
    return os.environ


def get_application_global() -> Optional[werkzeug.local.LocalProxy]:
    """
    Get the current application global proxy object.

    Returns
    -------
    proxy or None
    """
    if g:
        return g    # type: ignore
    return None



"""Runs docker login, build and push commands.

For use in travis ci.
"""

import os
import sys
from subprocess import call

try:
    tag = os.environ['TRAVIS_TAG']
    user = 'arxiv'
    context = sys.argv[1]
    label = sys.argv[2]
except Exception:
    print('Example of use:')
    print('TRAVIS_TAG=1.2.3 DOCKERHUB_PASSWORD=1234 DOCKERHUB_USERNAME=frank python -m arxiv.release.docker_build_push ./fourohfour fourohfour')
    exit(1)

try:
    labeltag = f"{user}/{label}:{tag}"
    login = 'docker login -u "$DOCKERHUB_USERNAME" -p "$DOCKERHUB_PASSWORD"'
    build = f"""docker build -t {labeltag} {context}"""
    push = f"""docker push {labeltag}"""
    logout = "docker logout"
    cmds = ' ' + ' && '.join([login, build, push, logout])

    # Running all the commands in a single subshell to keep docker login
    retcode = call(cmds, shell=True)
    if retcode < 0:
        print("Child was terminated by signal", -retcode, file=sys.stderr)
        exit(1)
    else:
        print("Child returned", retcode, file=sys.stderr)
        exit(retcode)
except OSError as e:
    print("Execution failed:", e, file=sys.stderr)
    exit(1)



"""Meta tools for integrations."""

import inspect
from typing import Any


class MetaIntegration(type):
    """
    Metaclass for context-bound integrations.

    The purpose of this metaclass is to simplify writing integrations that
    need to attach to the application or request context.

    A typical pattern  when using an integration's method is to first check for
    an instance of the integration on the current context, and then call the
    method on that instance. In practice, this means either performing the
    context check  every time the method is used (yuck) or implementing
    module-level functions that wrap the integration class instance methods,
    and check the context when called (also yuck).

    Here's an example of what life was like back then:


    .. code-block:: python

       class FileManagementService(object):
           '''Encapsulates a connection with the file management service.'''

           def __init__(self, *args, **kwargs) -> None:
               ...

           def get_upload_status(self, upload_id: int) -> Upload:
               ...

       def init_app(app: object = None) -> None:
           '''Set default configuration params for an application instance.'''
           config = get_application_config(app)
           config.setdefault('FILE_MANAGER_ENDPOINT', 'https://arxiv.org/')
           config.setdefault('FILE_MANAGER_VERIFY', True)

       def get_session(app: object = None) -> FileManagementService:
           '''Get a new session with the file management endpoint.'''
           config = get_application_config(app)
           endpoint = config.get('FILE_MANAGER_ENDPOINT', 'https://arxiv.org/')
           verify_cert = config.get('FILE_MANAGER_VERIFY', True)
           return FileManagementService(endpoint, verify_cert=verify_cert)

       def current_session() -> FileManagementService:
           '''Get/create :class:`.FileManagementService` for this context.'''
           g = get_application_global()
           if not g:
               return get_session()
           elif 'filemanager' not in g:
               g.filemanager = get_session()   # type: ignore
           return g.filemanager    # type: ignore

       @wraps(FileManagementService.get_upload_status)
       def get_upload_status(upload_id: int) -> Upload:
           '''See :meth:`FileManagementService.get_upload_status`.'''
           return current_session().get_upload_status(upload_id)


    and then when you want to use it:

    .. code-block:: python

       from my.cool.services import filemanager

       filemanager.get_upload_status(59192)


    A better way
    ------------
    Classes that use this metaclass are expected to implement a classmethod
    called ``current_session()`` that returns an instance of the class, similar
    to the example above. That method is responsible for getting the correct
    instance for the context.

    The net effect is that instance methods will be exposed as if they were
    class methods; behind the scenes, this metaclass will call
    ``current_session()`` and obtain the bound method from the returned
    instance.

    So you can do something like:

    .. code-block:: python

       class FileManagementService(metaclass=MetaIntegration):
           '''Encapsulates a connection with the file management service.'''

           def __init__(self, *args, **kwargs) -> None:
               ...

           def get_upload_status(self, upload_id: int) -> Upload:
               ...

           @classmethod
           def init_app(cls, app: object = None) -> None:
               '''Set default configuration params for an application instance.'''
               config = get_application_config(app)
               config.setdefault('FILE_MANAGER_ENDPOINT', 'https://arxiv.org/')
               config.setdefault('FILE_MANAGER_VERIFY', True)

           @classmethod
           def get_session(cls, app: object = None) -> 'FileManagementService':
               '''Get a new session with the file management endpoint.'''
               config = get_application_config(app)
               endpoint = config.get('FILE_MANAGER_ENDPOINT', 'https://arxiv.org/')
               verify_cert = config.get('FILE_MANAGER_VERIFY', True)
               return cls(endpoint, verify_cert=verify_cert)

           @classmethod
           def current_session(cls) -> 'FileManagementService':
               '''Get/create :class:`.FileManagementService` for this context.'''
               g = get_application_global()
               if not g:
                   return cls.get_session()
               elif 'filemanager' not in g:
                   g.filemanager = cls.get_session()   # type: ignore
               return g.filemanager    # type: ignore


    and then use it like:

    .. code-block:: python

       from my.cool.services.filemanager import FileManager

       FileManager.get_upload_status(59192)


    An even better way
    ------------------
    The example above was only a marginal improvement. What would be better
    is not having to write any of those classmethods at all.

    For an example of what that looks like, see
    :class:`.api.service.HTTPIntegration`.

    """

    def __getattribute__(self, key: str) -> Any:
        """Get the attribute from the instance bound to the current context."""
        obj = super(MetaIntegration, self).__getattribute__(key)
        if inspect.isfunction(obj):
            return getattr(self.current_session(), key)
        return obj



"""
Functions to deal with arxiv package versions.

It can be used in the setup.py file:
 from arxiv.release.dist_version import get_version
 setup(
     version=get_version('arxiv-filemanager'),
     ....
 )
"""
import sys
import pathlib
from subprocess import Popen, PIPE
from datetime import datetime
import pkg_resources
from pathlib import Path
from typing import Any, Optional


def get_version(dist_name: str) -> Optional[str]:
    """Get the version written by write_version(), or the git describe version.

    Parameters
    ----------
    dist_name: str
        Which arxiv distribution to get. ex arxiv-base
        arxiv-filemanager.  This should be the name from setup.py or
        pypi.  These will be mapped to arxiv.base.version and
        arxiv.filemanager.version.

    Returns
    -------
    str
        The version.__version__ value if it exists or the git describe
        version if it exists or the string 'no-git-or-release-version'

    """
    # TODO We might want to make it an error if we are under git
    # and there is a version.py file? It doesn't seem like a good state.
    pkg = ".".join(dist_name.split("-")) + ".version"
    try:
        name = "__version__"
        dist_version = str(getattr(__import__(pkg, fromlist=[name]), name))
        return dist_version
    except ModuleNotFoundError:
        pass

    pkv=get_pkg_version(dist_name)
    if pkv is not None:
        return pkv

    try:
        return get_git_version()
    except ValueError:
        pass
    return "0.0.1+no-git-or-release-version"


def write_version(dist_name: str, version: str) -> Path:
    """Write version to version.py in package corresponding with dist_name.

    Parameters
    ----------
    dist_name: str
        Which arxiv distribution to get. ex arxiv-base
        arxiv-filemanager.  These will be mapped to arxiv.base.version
        and arxiv.filemanager.version.

    version: str
        A string with a semantic version.

    Returns
    -------
    Path
        This returns the path to the version.py file.

    """
    dir = "/".join(dist_name.split("-")) + "/version.py"
    path = pathlib.Path(dir)
    path.parent.mkdir(parents=True, exist_ok=True)
    with open(path, "w+") as ff:  # overwrite existing version
        when = datetime.now().strftime("%d/%m/%Y %H:%M:%S")
        ff.write("'Created by tag_check.write_version'\n")
        ff.write("#  NEVER CHECK THIS version.py file INTO GIT.\n")
        ff.write(
            "#  Generated when the package was build for distribution.\n"
        )
        ff.write(f"__when__ = '{when}'\n")
        ff.write(f"__version__ = '{version}'\n")
    return path


def get_pkg_version(pkg: Any) -> Optional[str]:
    """Get the python package version.

    pkg needs to be the package name from setup.py or the name used to
    install from pypi.
    """
    try:
        return pkg_resources.get_distribution(pkg).version
    except:
        return None


def get_git_version(abbrev: int = 7) -> str:
    """Get the current version using `git describe`."""
    try:
        p = Popen(
            ["git", "describe", "--dirty", "--abbrev=%d" % abbrev],
            stdout=PIPE,
            stderr=PIPE,
        )
        p.stderr.close()
        line = p.stdout.readlines()[0]
        return str(line.strip().decode("utf-8"))
    except Exception:
        raise ValueError("Cannot get the version number from git")


# Below is intended to let this module be used in CI scripts:
# ``export APP_VER=$(python -m arxiv.release.get_version arxiv-hatsize-agent)``
if __name__ == "__main__":
    print(get_version(sys.argv[1]))



"""Exceptions raised by :class:`.BaseConsumer`."""


class CheckpointError(RuntimeError):
    """Checkpointing failed."""


class StreamNotAvailable(RuntimeError):
    """Could not find or connect to the stream."""


class KinesisRequestFailed(RuntimeError):
    """Raised when a Kinesis request failed permanently."""


class StopProcessing(RuntimeError):
    """Gracefully stopped processing upon unrecoverable error."""


class ConfigurationError(RuntimeError):
    """There was a problem with the configuration."""


class RestartProcessing(RuntimeError):
    """Something happened that requires a restart."""



'Created by tag_check.write_version'
#  NEVER CHECK THIS version.py file INTO GIT.
#  Generated when the package was build for distribution.
__when__ = '22/10/2020 15:07:21'
__version__ = '0.16.9rc3'



"""
Support for typed flash alerts.

Flask provides a `simple cookie-based flashing mechanism
<http://flask.pocoo.org/docs/1.0/patterns/flashing/>`. This module extends
that mechanism to support structured alerts (i.e. dicts) and set
message categories/severity.

.. note::

   For security purposes, alerts are not treated as safe in the template by
   default. To treat a message as safe, use ``safe=True`` when generating the
   flash notification. If you use ``safe=True`` you must escape any
   user input that goes into the alert.


For example:

.. code-block:: python

   from flask import url_for
   from arxiv.base import alerts
   from jinja2 import escape

   def handle_request():
       ...

       danger_txt = request.form.somefield.value
       help_url = url_for('help')
       alerts.flash_warning(f'This is a warning, see <a href="{help_url}">'
                            f'your input {escape(danger_txt)} was not found'
                            f'the docs</a> for more information',
                            title='Warning title', safe=True)
       alerts.flash_info('This is some info', title='Info title')
       alerts.flash_failure('This is a failure', title='Failure title')
       alerts.flash_success('This is a success', title='Success title')
       alerts.flash_warning('This is a warning that cannot be dismissed',
                            dismissable=False)


:func:`flash_hidden` can be used to send hidden data across requests.

"""

from arxiv.base import logging
from typing import Optional, List, Tuple, Union
from flask import flash, Markup, get_flashed_messages

INFO = 'info'
WARNING = 'warning'
FAILURE = 'danger'   # This is odd, but we use `danger` in styles.
SUCCESS = 'success'
HIDDEN = 'hidden'

logger = logging.getLogger(__name__)


def _flash_with(severity: str, message: Union[str, dict],
                title: Optional[str] = None, dismissable: bool = True,
                safe: bool = False) -> None:
    if safe and isinstance(message, str):
        message = Markup(message)
    data = {'message': message, 'title': title, 'dismissable': dismissable}
    logger.debug('flash with severity %s: (title: %s) %s',
                 severity, title, message)
    flash(data, severity)


def flash_info(message: str, title: Optional[str] = None,
               dismissable: bool = True, safe: bool = False) -> None:
    """
    Flash an informative message to the user.

    Parameters
    ----------
    message : str
        The content of the message.
    title : str or None
        An optional title, displayed prominently atop the notification.
    dismissable : bool
        If True, a button will be provided that allows the user to hide the
        notification.
    safe : bool
        If True, the message content (only) will be treated as safe for display
        in the HTML page. NB: only use this if you know for sure that the
        message content is safe for display (i.e. you're not using raw content
        from a request).

    """
    _flash_with(INFO, message, title, dismissable, safe)


def flash_warning(message: str, title: Optional[str] = None,
                  dismissable: bool = True, safe: bool = False) -> None:
    """
    Flash a warning message to the user.

    Warnings are like info alerts, but with a higher level of concern. For
    example, this might be used to notify a user that an unintended consequence
    might be imminent.

    Parameters
    ----------
    message : str
        The content of the message.
    title : str or None
        An optional title, displayed prominently atop the notification.
    dismissable : bool
        If True, a button will be provided that allows the user to hide the
        notification.
    safe : bool
        If True, the message content (only) will be treated as safe for display
        in the HTML page. NB: only use this if you know for sure that the
        message content is safe for display (i.e. you're not using raw content
        from a request).

    """
    _flash_with(WARNING, message, title, dismissable, safe)


def flash_failure(message: str, title: Optional[str] = None,
                  dismissable: bool = True, safe: bool = False) -> None:
    """
    Flash a failure message to the user.

    Failures are an unavoidable part of life. We should be honest about our
    mistakes, and find constructive ways to move forward with our lives. That
    means that we should provide some specific hints for how the user can
    achieve success (nb: this is an important accessibility requirement).

    Parameters
    ----------
    message : str
        The content of the message.
    title : str or None
        An optional title, displayed prominently atop the notification.
    dismissable : bool
        If True, a button will be provided that allows the user to hide the
        notification.
    safe : bool
        If True, the message content (only) will be treated as safe for display
        in the HTML page. NB: only use this if you know for sure that the
        message content is safe for display (i.e. you're not using raw content
        from a request).

    """
    _flash_with(FAILURE, message, title, dismissable, safe)


def flash_success(message: str, title: Optional[str] = None,
                  dismissable: bool = True, safe: bool = False) -> None:
    """
    Flash a success message to the user.

    It's nice to know when you've done something right for a change.

    Parameters
    ----------
    message : str
        The content of the message.
    title : str or None
        An optional title, displayed prominently atop the notification.
    dismissable : bool
        If True, a button will be provided that allows the user to hide the
        notification.
    safe : bool
        If True, the message content (only) will be treated as safe for display
        in the HTML page. NB: only use this if you know for sure that the
        message content is safe for display (i.e. you're not using raw content
        from a request).

    """
    _flash_with(SUCCESS, message, title, dismissable, safe)


def flash_hidden(message: dict, key: str,
                 dismissable: bool = True, safe: bool = False) -> None:
    """
    Propagate hidden data using the flash mechanism.

    Hidden messages are not shown to the user.

    Parameters
    ----------
    message : dict
        Data to flash across requests.
    key : str or None
        Key used to identify the data.
    dismissable : bool
        If True, a button will be provided that allows the user to hide the
        notification.
    safe : bool
        If True, the message content (only) will be treated as safe for display
        in the HTML page. NB: only use this if you know for sure that the
        message content is safe for display (i.e. you're not using raw content
        from a request).

    """
    _flash_with(HIDDEN, message, key, dismissable, safe)


def get_alerts(severity: Optional[str] = None) -> List[Tuple[str, dict]]:
    """
    Get displayable alerts.

    Parameters
    ----------
    severity : str or None
        If provided (default: None), only alerts of the specified severity
        will be loaded.

    Returns
    -------
    list
        Items are (str, dict) tuples, where the first element is the severity
        and the second element is the alert itself.

    """
    alerts: List[Tuple[str, dict]]
    if severity is not None:
        alerts = get_flashed_messages(with_categories=True,
                                      category_filter=[severity])
    else:
        alerts = get_flashed_messages(with_categories=True)
    return alerts


def get_hidden_alerts(key: str) -> Optional[dict]:
    """
    Get all hidden alerts.

    Parameters
    ----------
    key : str
        Key used to generate the hidden alert in the previous request.

    Returns
    -------
    dict or None

    """
    logger.debug('all current alerts: %s', get_flashed_messages())
    logger.debug('get hidden alert "%s"', key)
    messages = get_flashed_messages(category_filter=[HIDDEN])
    logger.debug('got messages: %s', messages)
    return {m['title']: m['message'] for m in messages}.get(key, None)



r"""
Patterns and functions to detect arXiv ids and URLs in text.

Functions to detect arXiv ids, URLs and DOI in text.
Functions to transform them to <a> tags.

These were originally jinja filters but became a little too big
for that so they were split out and made more general so they didn't
rely on the Flask context.

There are several classes of patterns we want to match but there is
some overlap in these patterns. To avoid looking for and parsing HTML in each
jinja filter, detecting these patterns is combined.

So far we are looking for:
DOIs            DOI: 10.1145/0001234.1234567
arXiv IDS:      1234.12345 1234.12345v1 hep-ph1307.1843
HTTP URLs:      http://something.org/myPaper/1234.12345
FTP URLs:       ftp://example.com/files/1234.12345

Just matching for arXiv ids with \d{4}\.\d{4,5} will match several of
these. To deal with this we are prioritizing the matches and
interrupting once one is found.

We should probably match DOIs first because they are the source of a
lot of false positives for arxiv matches.

Updated 18 January, 2019: refactored to combine arXiv + DOI regexes with URL
matching and link generation provided by `bleach
<https://bleach.readthedocs.io>`_. The bleach URL regex is extended a bit to
handle FTP addresses.

Updated 20 March, 2019: refactored to provide independent bleach attribute
callbacks for each kind of link.
"""
from typing import List, Pattern, Tuple, Callable, Dict, Union, Text, Any, \
    MutableMapping
import re
from functools import reduce
from urllib.parse import quote, urlparse

from flask import url_for, g
import bleach

from arxiv.taxonomy import CATEGORIES
from arxiv import identifier
from . import clickthrough


Attrs = MutableMapping[Any, Text]
Callback = Callable[[Attrs, bool], Attrs]
Callable_Linker = Callable[[str], str]


def _without_group_names(pattern: Pattern) -> str:
    ptn: str = re.sub(r'\(\?P<[^>\!]+>', '(?:', pattern.pattern)
    return ptn


DOI = re.compile(   # '10.1145/0001234.1234567'
    r'(?P<doi>10.\d{4,9}/[-._;()/:A-Z0-9]+)',
    re.I
)
"""
Pattern for matching DOIs.

We should probably match DOIs first because they are the source of a
lot of false positives for arxiv id matches.

Only using the most general expression from
https://www.crossref.org/blog/dois-and-matching-regular-expressions/
"""


BROAD_DOI = re.compile(r'(?P<doi>\d{2,3}.\d{4,5}\/\S+)', re.I)
"""
Very broad pattern for matching DOIs in the abs DOI field.

Ex. 10.1175/1520-0469(1996)053<0946:ASTFHH>2.0.CO;2
Ex. 21.11130/00-1735-0000-0005-146A-E
"""

ARXIV_PATTERNS = [
    identifier.OLD_STYLE_WITH_ARCHIVE,  # math/0501233 hep-ph/0611734
    identifier.OLD_STYLE_WITH_CATEGORY,  # math.GR/0601136v3 math.GR/0601136
    identifier.STANDARD  # 1609.05068 1207.1234v1 1207.1234 1807.12345v12
]
ARXIV_ID = re.compile(
    "|".join([rf"(?:{_without_group_names(pattern)})"
              for pattern in ARXIV_PATTERNS]),
    re.IGNORECASE | re.VERBOSE | re.UNICODE
)

OKCHARS = r'([a-z0-9,_.\-+~:]|%[a-f0-9]*)'
"""Characters that are acceptable during PATH, QUERY and ANCHOR parts"""

PATH = rf'(?P<PATH>(/{OKCHARS}*)+)?'
"""Regex for path part of URLs for use in urlize"""

FTP = re.compile(rf'(?P<url>(?:ftp://)({OKCHARS}|(@))*{PATH})', re.I)
"""Regex to match FTP URLs in text."""

IP_ADDRESS = (r'(?:(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\.){3}'
              r'(?:25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)')
"""Regex to match an IP address."""

TLDS = '|'.join(bleach.linkifier.TLDS)
PROTOCOLS = '|'.join(bleach.linkifier.html5lib_shim.allowed_protocols)
URL = re.compile(
    rf"""(?:{FTP.pattern})|
    (?:\(*  # Match any opening parentheses.
    \b(?<![@/])(?:(?:{PROTOCOLS}):/{{0,3}}(?:(?:\w+:)?\w+@)?)?  # http://
    (?:
        (?:([\w-]+\.)+(?:{TLDS}))   # Conventional URL
        |(?:{IP_ADDRESS})           # IP address
    )(?:\:[0-9]+)?(?!\.\w)\b        # Port
    (?:[/?][^\s\{{\}}\|\\\^\[\]`<>"]*)?)
        # /path/zz (excluding "unsafe" chars from RFC 1738,
        # except for # and ~, which happen in practice)
    """,
    re.IGNORECASE | re.VERBOSE | re.UNICODE
)


def arxiv_id_to_url(arxiv_id: str) -> str:
    """Generate an URL for an arXiv ID."""
    url: str = url_for('abs_by_id', paper_id=arxiv_id)
    return url


def url_for_doi(doi: str) -> str:
    """Generate an URL for a DOI."""
    quoted_doi = quote(doi, safe='/')
    return f'https://dx.doi.org/{quoted_doi}'


def clickthrough_url_for_doi(doi: str) -> str:
    """Generate a clickthrough URL for a DOI."""
    return clickthrough.clickthrough_url(url_for_doi(doi))


def _extend_class_attr(attrs: Attrs, new_class: str) -> Attrs:
    if (None, 'class') not in attrs:
        attrs[(None, 'class')] = ''
    attrs[(None, 'class')] = (attrs[(None, 'class')] + f' {new_class}').strip()
    return attrs


def _add_rel(attrs: Attrs, new: bool = False) -> Attrs:
    """Check if href is internal or external and adds attrs as appropriate."""
    o = urlparse(attrs[(None, 'href')])
    if not o.netloc.split(':')[0].endswith('arxiv.org'):   # External link?
        return _add_rel_external(attrs, new)
    else:
        return _add_rel_internal(attrs, new)


def _add_rel_external(attrs: Attrs, new: bool = False) -> Attrs:
    """Add an external rel."""
    # noopener for security reasons
    # nofollow to disincentivize arXiv articles for SEO
    # external says that link is away from arxiv
    attrs[(None, 'rel')] = 'external noopener nofollow'
    _extend_class_attr(attrs, 'link-external')
    return attrs


def _add_rel_internal(attrs: Attrs, new: bool = False) -> Attrs:
    """Add an internal rel."""
    _extend_class_attr(attrs, 'link-internal')
    return attrs


def _this_url_text(attrs: Attrs, new: bool = False) -> Attrs:
    o = urlparse(attrs[(None, 'href')])
    if o.hostname and not o.hostname.endswith('arxiv.org'):
        attrs['_text'] = f'this {o.scheme} URL'    # Replaces the link text.
    return attrs


def _add_scheme_info(attrs: Attrs, new: bool = False) -> Attrs:
    o = urlparse(attrs[(None, 'href')])
    if (None, 'class') not in attrs:
        attrs[(None, 'class')] = ''
    _extend_class_attr(attrs, f'link-{o.scheme}')
    return attrs


def _handle_arxiv_url(attrs: Attrs, new: bool = False) -> Attrs:
    """
    Screen for reference to an arXiv e-print, and generate URL.

    If the :ref:`.ARXIV_ID` pattern is used, it will generate a link with the
    identifier as a the target. We need to intercept these links, and generate
    a real URL.
    """
    href = attrs[(None, 'href')]
    if '://' in href:   # http://1902.00123
        target = href.split('://', 1)[1]
        prefix = ''
    elif ':' in href:   # arxiv:1902.00123
        target = href.split(':', 1)[1]
        prefix = 'arXiv:'
    else:
        target = href
        prefix = ''
    if ARXIV_ID.match(target):
        attrs[(None, 'href')] = arxiv_id_to_url(target)
        attrs['_text'] = f'{prefix}{target}'
        attrs[(None, 'data-arxiv-id')] = target     # Add arxiv="<arxiv id>"
    return attrs


def _handle_doi_url(attrs: Attrs, new: bool = False) -> Attrs:
    """
    Screen for reference to a DOI, and generate a URL.

    If the :ref:`.DOI` pattern is used, it will generate a link with the
    doi as a the target. We need to intercept these links, and generate
    a real URL.
    """
    href = attrs[(None, 'href')]
    if '://' in href:
        target = href.split('://', 1)[1]
        prefix = ''
    elif ':' in href:
        target = href.split(':', 1)[1]
        prefix = 'doi:'
    else:
        target = href
        prefix = ''
    if DOI.match(target):
        attrs[(None, 'href')] = clickthrough_url_for_doi(target)
        attrs['_text'] = f'{prefix}{target}'
        attrs[(None, 'data-doi')] = target     # Add arxiv="<arxiv id>"
    return attrs


def _handle_broad_doi_url(attrs: Attrs, new: bool = False) -> Attrs:
    """
    Handle doi from DOI field.

    It is always just a DOI so turn it into a DOI link.
    """
    doi_attrs = _handle_doi_url(attrs, new)
    if (None, 'data-doi') in doi_attrs:
        return doi_attrs
    else:
        target = attrs['_text']
        attrs[(None, 'href')] = clickthrough_url_for_doi(target)
        attrs[(None, 'data-doi')] = target
        return attrs


ENDS_WITH_TLD = r".*\.(" + TLDS + ")$"
CATEGORIES_THAT_COULD_BE_HOSTNAMES = '|'.join([cat_id for cat_id in CATEGORIES.keys()
                                               if re.search(ENDS_WITH_TLD, cat_id, re.IGNORECASE)])
DONT_URLIZE_CATS = re.compile(CATEGORIES_THAT_COULD_BE_HOSTNAMES,
                              re.IGNORECASE | re.UNICODE)


def _dont_urlize_arxiv_categories(attrs: Attrs, new: bool = False) -> Attrs:
    """
    Prevent urlizing archive categories that look like hostnames.

    Ex. don't urlize math.CO but do urlize supermath.co
    """
    url = urlparse(attrs[(None, 'href')])
    if DONT_URLIZE_CATS.match(url.netloc):
        return None  # type: ignore
    else:
        return attrs


PATTERNS = {
    'doi': DOI,
    'doi_field': BROAD_DOI,
    'url': URL,
    'arxiv_id': ARXIV_ID
}
ORDER = ['arxiv_id', 'doi', 'url', 'doi_field']
DEFAULT_KINDS = ['arxiv_id', 'doi', 'url']
"""Default list of identifier types to match and convert to to URLs.

This does not include 'doi_field because that is a specialized kind
for just the abs DOI field."""

callbacks = {
    'doi':      [_handle_doi_url, _add_scheme_info, _add_rel_external],
    'doi_field': [_handle_broad_doi_url, _add_scheme_info, _add_rel_external],
    'arxiv_id': [_handle_arxiv_url, _add_scheme_info],
    'url':      [_dont_urlize_arxiv_categories, _this_url_text, _add_rel, _add_scheme_info]
}
"""Bleach attribute callbacks for each kind."""


def _get_pattern(kinds: List[str]) -> Pattern:
    return re.compile(
        '|'.join([rf'(?:{PATTERNS[kind].pattern})' for kind
                  in sorted(kinds, key=lambda kind: ORDER.index(kind))]),
        re.IGNORECASE | re.VERBOSE | re.UNICODE
    )

def _get_linker_of_kind(kind: str) -> Callable_Linker:
    # The returned object will attempt to match url_re in the input string.
    # For each url_re that matches, it will run all of the attribute
    # adjusting callbacks and then stick the tag in the output string
    # replacing the matched text.
    linker: Callable[[str], str] = bleach.linkifier.Linker(
        callbacks=callbacks[kind],
        skip_tags=['a'],
        url_re=_get_pattern([kind])).linkify
    return linker


def _compose_list_of_funcs(fv: List[Callable_Linker]) -> Callable_Linker:
    """Return function that calls fv functions one after another."""
    if not fv:
        return lambda x: x
    return reduce(lambda f, g: lambda x: f(g(x)), fv[1:], fv[0])


def _deferred_thread_local_linker_of_kind(kind: str) -> Callable_Linker:
    # Bleach linkifiers are not thread safe. This puts them on Flask.g.
    #
    # This will return a Callable that creates the bleach linkifier
    # only when called with the string to linkify. This is so the
    # bleach linkifier can be put into a Flask.g of each
    # individual WSGI thread.

    def deferred(instr: str) -> str:
        # This must not be called while the app is starting up, only
        # when the linkiers are being used in requests. So the
        # function is created/cached at call time.
        if 'linkers' not in g:
            g.linkers = {}
        if kind not in g.linkers:
            g.linkers[kind] = _get_linker_of_kind(kind)

        return g.linkers[kind](instr) #type: ignore

    return deferred


def _get_linker(kinds: List[str]) -> Callable_Linker:
    # Gets composed, thread-local, deferred linkers.
    # This is intended to be run as the flask app starts up
    # or in the loading of packages.
    # The actual creating of thread local bleach objects is
    # done in _deferred_thread_local_linker_of_kind.
    if len(kinds) > 1 and 'doi_field' in kinds:
        raise ValueError(
            'doi_field should not be used in combination with other kinds')
    ordered_linkers = [_deferred_thread_local_linker_of_kind(kind) for kind
                       in sorted(kinds, key=lambda kind: ORDER.index(kind))]
    return _compose_list_of_funcs(ordered_linkers)


def urlize(text: str, kinds: List[str] = DEFAULT_KINDS) -> str:
    """
    Convert URLs and certain identifiers to HTML links.

    Parameters
    ----------
    text : str
        Some text containing identifiers and/or URLs.
    kinds : list
        Names (from :ref:`SUPPORTED_KINDS`) of identifiers to match.

    Returns
    -------
    str
        The passed text, with identifiers and/or URLs converted to HTML links.

    """
    return _get_linker(kinds)(text)


def urlizer(kinds: List[str] = DEFAULT_KINDS) -> Callable_Linker:
    """Generate a function to convert tokens to links.

    If the urlizing function is going to be reused, this is more
    efficient than urlize because this will only call _get_linker()
    once. urlize() will call _get_linker() on each call to urlize(txt)

    Parameters
    ----------
    kinds : list
        Names (from :ref:`SUPPORTED_KINDS`) of identifiers to match and
        convert.

    Returns
    -------
    callable
        A function with signature ``(str) -> str`` that converts identifiers
        to HTML links.

    """
    return _get_linker(kinds)



"""Convert between TeX escapes and UTF8."""
import re
from typing import Pattern, Dict, Match

accents = {
    # first accents with non-letter prefix, e.g. \'A
    "'A": 0x00c1, "'C": 0x0106, "'E": 0x00c9, "'I": 0x00cd,
    "'L": 0x0139, "'N": 0x0143, "'O": 0x00d3, "'R": 0x0154,
    "'S": 0x015a, "'U": 0x00da, "'Y": 0x00dd, "'Z": 0x0179,
    "'a": 0x00e1, "'c": 0x0107, "'e": 0x00e9, "'i": 0x00ed,
    "'l": 0x013a, "'n": 0x0144, "'o": 0x00f3, "'r": 0x0155,
    "'s": 0x015b, "'u": 0x00fa, "'y": 0x00fd, "'z": 0x017a,
    '"A': 0x00c4, '"E': 0x00cb, '"I': 0x00cf, '"O': 0x00d6,
    '"U': 0x00dc, '"Y': 0x0178, '"a': 0x00e4, '"e': 0x00eb,
    '"i': 0x00ef, '"o': 0x00f6, '"u': 0x00fc, '"y': 0x00ff,
    '.A': 0x0226, '.C': 0x010a, '.E': 0x0116, '.G': 0x0120,
    '.I': 0x0130, '.O': 0x022e, '.Z': 0x017b, '.a': 0x0227,
    '.c': 0x010b, '.e': 0x0117, '.g': 0x0121, '.o': 0x022f,
    '.z': 0x017c, '=A': 0x0100, '=E': 0x0112, '=I': 0x012a,
    '=O': 0x014c, '=U': 0x016a, '=Y': 0x0232, '=a': 0x0101,
    '=e': 0x0113, '=i': 0x012b, '=o': 0x014d, '=u': 0x016b,
    '=y': 0x0233, '^A': 0x00c2, '^C': 0x0108, '^E': 0x00ca,
    '^G': 0x011c, '^H': 0x0124, '^I': 0x00ce, '^J': 0x0134,
    '^O': 0x00d4, '^S': 0x015c, '^U': 0x00db, '^W': 0x0174,
    '^Y': 0x0176, '^a': 0x00e2, '^c': 0x0109, '^e': 0x00ea,
    '^g': 0x011d, '^h': 0x0125, '^i': 0x00ee, '^j': 0x0135,
    '^o': 0x00f4, '^s': 0x015d, '^u': 0x00fb, '^w': 0x0175,
    '^y': 0x0177, '`A': 0x00c0, '`E': 0x00c8, '`I': 0x00cc,
    '`O': 0x00d2, '`U': 0x00d9, '`a': 0x00e0, '`e': 0x00e8,
    '`i': 0x00ec, '`o': 0x00f2, '`u': 0x00f9, '~A': 0x00c3,
    '~I': 0x0128, '~N': 0x00d1, '~O': 0x00d5, '~U': 0x0168,
    '~a': 0x00e3, '~i': 0x0129, '~n': 0x00f1, '~o': 0x00f5,
    '~u': 0x0169,
    # and now ones with letter prefix \c{c} etc..
    'HO': 0x0150, 'HU': 0x0170, 'Ho': 0x0151, 'Hu': 0x0171,
    'cC': 0x00c7, 'cE': 0x0228,
    'cG': 0x0122, 'cK': 0x0136, 'cL': 0x013b, 'cN': 0x0145,
    'cR': 0x0156, 'cS': 0x015e, 'cT': 0x0162, 'cc': 0x00e7,
    'ce': 0x0229, 'cg': 0x0123, 'ck': 0x0137, 'cl': 0x013c,
    # Commented out due ARXIVDEV-2322 (bug reported by PG)
    # 'ci' : 'i\x{0327}' = chr(0x69).ch(0x327) # i with combining cedilla
    'cn': 0x0146, 'cr': 0x0157, 'cs': 0x015f, 'ct': 0x0163,
    'kA': 0x0104, 'kE': 0x0118, 'kI': 0x012e, 'kO': 0x01ea,
    'kU': 0x0172, 'ka': 0x0105, 'ke': 0x0119, 'ki': 0x012f,
    'ko': 0x01eb, 'ku': 0x0173, 'rA': 0x00c5, 'rU': 0x016e,
    'ra': 0x00e5, 'ru': 0x016f, 'uA': 0x0102, 'uE': 0x0114,
    'uG': 0x011e, 'uI': 0x012c, 'uO': 0x014e, 'uU': 0x016c,
    'ua': 0x0103, 'ue': 0x0115, 'ug': 0x011f,
    'ui': 0x012d, 'uo': 0x014f, 'uu': 0x016d,
    'vA': 0x01cd, 'vC': 0x010c, 'vD': 0x010e,
    'vE': 0x011a, 'vG': 0x01e6, 'vH': 0x021e, 'vI': 0x01cf,
    'vK': 0x01e8, 'vL': 0x013d, 'vN': 0x0147, 'vO': 0x01d1,
    'vR': 0x0158, 'vS': 0x0160, 'vT': 0x0164, 'vU': 0x01d3,
    'vZ': 0x017d, 'va': 0x01ce, 'vc': 0x010d, 'vd': 0x010f,
    've': 0x011b, 'vg': 0x01e7, 'vh': 0x021f, 'vi': 0x01d0,
    'vk': 0x01e9, 'vl': 0x013e, 'vn': 0x0148, 'vo': 0x01d2,
    'vr': 0x0159, 'vs': 0x0161, 'vt': 0x0165, 'vu': 0x01d4,
    'vz': 0x017e
}
r"""
Hash to lookup tex markup and convert to Unicode.

macron: a line above character (overbar \={} in TeX)
caron: v-shape above character (\v{ } in TeX)
See: http://www.unicode.org/charts/

"""

textlet = {
    'AA': 0x00c5, 'AE': 0x00c6, 'DH': 0x00d0, 'DJ': 0x0110,
    'ETH': 0x00d0, 'L': 0x0141, 'NG': 0x014a, 'O': 0x00d8,
    'oe': 0x0153, 'OE': 0x0152, 'TH': 0x00de, 'aa': 0x00e5,
    'ae': 0x00e6,
    'dh': 0x00f0, 'dj': 0x0111, 'eth': 0x00f0, 'i': 0x0131,
    'l': 0x0142, 'ng': 0x014b, 'o': 0x00f8, 'ss': 0x00df,
    'th': 0x00fe,
    }

textgreek = {
    # Greek (upper)
    'Gamma': 0x0393, 'Delta': 0x0394, 'Theta': 0x0398,
    'Lambda': 0x039b, 'Xi': 0x039E, 'Pi': 0x03a0,
    'Sigma': 0x03a3, 'Upsilon': 0x03a5, 'Phi': 0x03a6,
    'Psi': 0x03a8, 'Omega': 0x03a9,
    # Greek (lower)
    'alpha': 0x03b1, 'beta': 0x03b2, 'gamma': 0x03b3,
    'delta': 0x03b4, 'epsilon': 0x03b5, 'zeta': 0x03b6,
    'eta': 0x03b7, 'theta': 0x03b8, 'iota': 0x03b9,
    'kappa': 0x03ba, 'lambda': 0x03bb, 'mu': 0x03bc,
    'nu': 0x03bd, 'xi': 0x03be, 'omicron': 0x03bf,
    'pi': 0x03c0, 'rho': 0x03c1, 'varsigma': 0x03c2,
    'sigma': 0x03c3, 'tau': 0x03c4, 'upsion': 0x03c5,
    'varphi': 0x03C6,  # φ
    'phi':  0x03D5,  # ϕ
    'chi': 0x03c7, 'psi': 0x03c8, 'omega': 0x03c9,
}


def _p_to_match(tex_to_chr: Dict[str, int]) -> Pattern:
    # textsym and textlet both use the same sort of regex pattern.
    keys = r'\\(' + '|'.join(tex_to_chr.keys()) + ')'
    pstr = r'({)?' + keys + r'(\b|(?=_))(?(1)}|(\\(?= )| |{}|)?)'
    return re.compile(pstr)


textlet_pattern = _p_to_match(textlet)
textgreek_pattern = _p_to_match(textgreek)

textsym = {
    'P': 0x00b6, 'S': 0x00a7, 'copyright': 0x00a9,
    'guillemotleft': 0x00ab, 'guillemotright': 0x00bb,
    'pounds': 0x00a3, 'dag': 0x2020, 'ddag': 0x2021,
    'div': 0x00f7, 'deg': 0x00b0}

textsym_pattern = _p_to_match(textsym)


def _textlet_sub(match: Match) -> str:
    return chr(textlet[match.group(2)])


def _textsym_sub(match: Match) -> str:
    return chr(textsym[match.group(2)])


def _textgreek_sub(match: Match) -> str:
    return chr(textgreek[match.group(2)])


def texch2UTF(acc: str) -> str:
    """Convert single character TeX accents to UTF-8.

    Strip non-whitepsace characters from any sequence not recognized (hence
    could return an empty string if there are no word characters in the input
    string).

    chr(num) will automatically create a UTF8 string for big num
    """
    if acc in accents:
        return chr(accents[acc])
    else:
        return re.sub(r'[^\w]+', '', acc, flags=re.IGNORECASE)


def tex2utf(tex: str, greek: bool = True) -> str:
    r"""Convert some TeX accents and greek symbols to UTF-8 characters.

    :param tex: Text to filter.

    :param greek: If False, do not convert greek letters or
    ligatures.  Greek symbols can cause problems. Ex. \phi is not
    suppose to look like φ. φ looks like \varphi.  See ARXIVNG-1612

    :returns: string, possibly with some TeX replaced with UTF8

    """
    # Do dotless i,j -> plain i,j where they are part of an accented i or j
    utf = re.sub(r"/(\\['`\^\"\~\=\.uvH])\{\\([ij])\}", r"\g<1>\{\g<2>\}", tex)

    # Now work on the Tex sequences, first those with letters only match
    utf = textlet_pattern.sub(_textlet_sub, utf)

    if greek:
        utf = textgreek_pattern.sub(_textgreek_sub, utf)

    utf = textsym_pattern.sub(_textsym_sub, utf)

    utf = re.sub(r'\{\\j\}|\\j\s', 'j', utf)  # not in Unicode?

    # reduce {{x}}, {{{x}}}, ... down to {x}
    while re.search(r'\{\{([^\}]*)\}\}', utf):
        utf = re.sub(r'\{\{([^\}]*)\}\}', r'{\g<1>}', utf)

    # Accents which have a non-letter prefix in TeX, first \'e
    utf = re.sub(r'\\([\'`^"~=.][a-zA-Z])',
                 lambda m: texch2UTF(m.group(1)), utf)

    # then \'{e} form:
    utf = re.sub(r'\\([\'`^"~=.])\{([a-zA-Z])\}',
                 lambda m: texch2UTF(m.group(1) + m.group(2)), utf)

    # Accents which have a letter prefix in TeX
    #  \u{x} u above (breve), \v{x}   v above (caron), \H{x}   double accute...
    utf = re.sub(r'\\([Hckoruv])\{([a-zA-Z])\}',
                 lambda m: texch2UTF(m.group(1) + m.group(2)), utf)

    # Don't do \t{oo} yet,
    utf = re.sub(r'\\t{([^\}])\}', r'\g<1>', utf)

    # bdc34: commented out in original Perl
    # $utf =~ s/\{(.)\}/$1/g; #  remove { } from around {x}

    return utf



"""Template filters."""

import re
import os
from functools import partial
from datetime import datetime
from pytz import timezone
from typing import Union

from flask import Flask
from jinja2 import Markup, escape

from arxiv.base.urls import urlizer, canonical_url, clickthrough_url
from arxiv.util.tex2utf import tex2utf
from arxiv.taxonomy import get_category_display, get_archive_display, \
    get_group_display

ET = timezone('US/Eastern')


JinjaFilterInput = Union[Markup, str]
"""
   Jinja filters will receive their text input as either
   a Markup object or a str. It is critical for proper escaping to
   to ensure that str is correctly HTML escaped.

   Markup is decoded from str so this type is redundant but
   the hope is to make it clear what is going on to arXiv developers.
"""


def abstract_lf_to_br(text: JinjaFilterInput) -> Markup:
    """Lines that start with two spaces should be broken."""
    if isinstance(text, Markup):
        etxt = text
    else:
        etxt = Markup(escape(text))

    # if line starts with spaces, replace the white space with <br />
    br = re.sub(r'((?<!^)\n +)', '\n<br />', etxt)
    dedup = re.sub(r'\n\n', '\n', br)  # skip if blank
    return Markup(dedup)


def f_tex2utf(text: JinjaFilterInput,
              greek: bool = True) -> Markup:
    """Return output of tex2utf function as escaped Markup."""
    if isinstance(text, Markup):
        return escape(tex2utf(text.unescape(), greek=greek))
    else:
        return Markup(escape(tex2utf(text, greek=greek)))


def embed_content(path: str) -> Markup:
    """Embed the content of a static file."""
    stat = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'static')
    with open(os.path.join(stat, path)) as f:
        return Markup(f.read())


def tidy_filesize(size: Union[int, float]) -> str:
    """
    Convert upload size to human readable form.

    Decision to use powers of 10 rather than powers of 2 to stay compatible
    with Jinja filesizeformat filter with binary=false setting that we are
    using in file_upload template.

    Parameter: size in bytes
    Returns: formatted string of size in units up through GB

    """
    units = ["B", "KB", "MB", "GB"]
    if size == 0:
        return "0B"
    if size > 1000000000:
        return '{} {}'.format(size, units[3])
    units_index = 0
    while size > 1000:
        units_index += 1
        size = round(size / 1000, 3)
    return '{} {}'.format(size, units[units_index])


def as_eastern(utc_datetime: datetime) -> datetime:
    """Relocalize an UTC datetime in US/Eastern time."""
    return utc_datetime.astimezone(ET)


def register_filters(app: Flask) -> None:
    """
    Register base template filters on a Flask app.

    Parameters
    ----------
    app : :class:`Flask`

    """
    app.template_filter('abstract_lf_to_br')(abstract_lf_to_br)
    app.template_filter('urlize')(urlizer())
    app.template_filter('tex2utf')(partial(f_tex2utf, greek=True))
    app.template_filter('tex2utf_no_symbols')(partial(f_tex2utf, greek=False))
    app.template_filter('canonical_url')(canonical_url)
    app.template_filter('clickthrough_url')(clickthrough_url)
    app.template_filter('get_category_display')(get_category_display)
    app.template_filter('get_archive_display')(get_archive_display)
    app.template_filter('get_group_display')(get_group_display)
    app.template_filter('embed_content')(embed_content)
    app.template_filter('tidy_filesize')(tidy_filesize)
    app.template_filter('as_eastern')(as_eastern)
    app.template_filter('abs_doi_to_urls')(urlizer(['doi_field']))
    app.template_filter('arxiv_id_urlize')(urlizer(['arxiv_id']))



"""Install arXiv-base as an importable package.

This needs to be able to install from a directory that has a
version.py file or from a directory that has no version file but does
have git

To test with no version but with git:
```
$ cd tmp
$ git clone git@github.com:arXiv/arxiv-base.git
$ cd arxiv-base
$ pip install ./
...
$ pip show arxiv-base | grep Version
Version: 0.16.6.2-dev-3-ga5af154 
```

To test with a version file but no git:
```
$ cd tmp
$ git clone git@github.com:arXiv/arxiv-base.git
$ cd arxiv-base
$ rm -rf .git
$ python 
Python 3.6.9 (default, Dec 23 2019, 12:56:30)[GCC 7.4.0] on linux
>>> import arxiv.release.dist_version
>>> arxiv.release.dist_version.write_version('arxiv-base', '10.9.9')
CTRL-D
$ pip install ./
...
$ pip show arxiv-base | grep Version
Version: 10.9.9
```

"""

import codecs
import os
import sys
import re
from subprocess import Popen, PIPE

from setuptools import setup, find_packages


here = os.path.abspath(os.path.dirname(__file__))

def read(*parts):
    with codecs.open(os.path.join(here, *parts), 'r') as fp:
        return fp.read()

def find_version(*file_paths):
    try:
        version_file = read(*file_paths)
        version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
        if version_match:
            return version_match.group(1)
    except Exception:
        pass # no file is find, just try to get it from git

    try:
        p = Popen(['git', 'describe', '--dirty'],
                  stdout=PIPE, stderr=PIPE)
        p.stderr.close()
        line = p.stdout.readlines()[0]
        return line.strip().decode('utf-8')
    except Exception:
        raise RuntimeError("Unable to find version string in version file or git. See setup.py for instructions.")


setup(
    name='arxiv-base',
    version=find_version('arxiv', 'base', 'version.py'),
    packages=[f'arxiv.{package}' for package
              in find_packages('arxiv', exclude=['*test*'])],
    zip_safe=False,
    install_requires=[
        'flask',
        'jsonschema',
        'pytz',
        'uwsgi',
        'boto3',
        'bleach==3.1.0',
        'backports-datetime-fromisoformat==1.0.0',
        'typing-extensions'
    ],
    include_package_data=True
)



"""Functions to ensure git tag version is correct for travis and to get ver.

This is intended as a check to use in travis ci.

This will look to see if there is a TRAVIS_TAG env var,
if there is it means we have a tag in the CI system.

It will check that tag to see if it is a valid public python version
that pypi will accept.

If it is valid, it will make a version.py file.

If there is no TRAVIS_TAG or it is invalid, this will do nothing.

Invalid tags should be prevented from getting into pypi by pypi
just rejecting the upload.

It can be used in a travis.yml:
script:
- pip install arxiv-base
- python deploy/prepare_for_version.py


This can also be used to check if a version string is valid:
``TRAVIS_TAG=0.2.3-somethingX python -m arxiv.release.tag_check arxiv-base``

This does not write or update a RELEASE-VERSION file.
"""


from typing import Any
import sys
import os
import re
from .dist_version import write_version

NO_TAG_MSG = (
    "OK: Skipping publish version check since no git tag found in TRAVIS_TAG"
)

REGRESSIVE_MSG = "NOT OK: Tag did not pass tag check"
INVALID_MSG = "OK: skipping publish version since the tag is not a valid PEP440 public python version. See https://www.python.org/dev/peps/pep-0440/#version-scheme"


def prepare_for_version(dist_name: str) -> Any:
    """Prepare for a version on travis-ci.

    Intended to be used when prepareing a version on travis-ci or
    other CI. This will check if a tag exists and verify it is
    good. Then it will write the version to a python file that will be
    used by both the app code and setup.py.

    This does not check if the tag is redundent since by the time travis
    runs anything, the tag will already be in git.
    """
    tag_to_publish = os.environ.get("TRAVIS_TAG", None)

    if not tag_to_publish:
        print(NO_TAG_MSG)
        return 0

    if not is_valid_python_public_version(tag_to_publish):
        print(INVALID_MSG)
        return 0

    topkg = write_version(dist_name, tag_to_publish)
    print(f"Wrote version {tag_to_publish} to {topkg}")


def is_valid_python_public_version(tag: str) -> bool:
    """Check if tag is valid PEP440 public python version."""
    return bool(_pep440public_ver.match(tag))


# Patern from PEP440 with local removed
VERSION_PATTERN = r"""
    v?
    (?:
        (?:(?P<epoch>[0-9]+)!)?                           # epoch
        (?P<release>[0-9]+(?:\.[0-9]+)*)                  # release segment
        (?P<pre>                                          # pre-release
            [-_\.]?
            (?P<pre_l>(a|b|c|rc|alpha|beta|pre|preview))
            [-_\.]?
            (?P<pre_n>[0-9]+)?
        )?
        (?P<post>                                         # post release
            (?:-(?P<post_n1>[0-9]+))
            |
            (?:
                [-_\.]?
                (?P<post_l>post|rev|r)
                [-_\.]?
                (?P<post_n2>[0-9]+)?
            )
        )?
        (?P<dev>                                          # dev release
            [-_\.]?
            (?P<dev_l>dev)
            [-_\.]?
            (?P<dev_n>[0-9]+)?
        )?
    )
"""

_pep440public_ver = re.compile(
    r"^\s*" + VERSION_PATTERN + r"\s*$", re.VERBOSE | re.IGNORECASE,
)

if __name__ == "__main__":
    """ This is intended to let this module be used in CI scripts:
    ``python -m arxiv.release.tag_check``
    """
    prepare_for_version(sys.argv[1])



"""
Helpers for working with URLs in arXiv Flask applications.

This module provides :func:`external_url_handler`, which is attached to Flask
application instances by :class:`arxiv.base.Base`. This leverage's Flask's
built-in hook for :class:`flask.BuildError` exception handlers, described in
the :func:`flask.url_for` documentation.

To set a URL for an external (i.e. not provided by thie application) endpoint,
define a variable called ``URLS`` in your application configuration. ``URLS``
should be a list of three-tuples, of the form (``endpoint (str)``,
``path (str)``, ``domain (str)``). Paths should use the `Werkzeug rule format
<http://werkzeug.pocoo.org/docs/0.14/routing/#rule-format>`_. For example:

.. code-block:: python

   URLS = [
       ("pdf", "/pdf/<arxiv:paper_id>", "arxiv.org"),
       ("twitter", "/arxiv", "twitter.com"),
       ("blog", "/arxiv", "blogs.cornell.edu")
   ]

You can load these urls using :func:`flask.url_for`. For example,
``url_for("pdf", paper_id="1601.00123")`` should return
``https://arxiv.org/pdf/1601.00123``.

Using environment variables
---------------------------
An alternative approach is to set the endpoint using an environment variable.
Setting ``ARXIV_FOO_URL`` will define a rule for endpoint ``foo`` using the
domain and path parsed from the variable's value. For example:

.. code-block:: python

   os.environ['ARXIV_BLOG_URL'] = 'http://blogs.cornell.edu/arxiv'


Caveats
-------
URLs can be defined in :mod:`arxiv.base.config`, in environment variables,
and in the configuration of a :class:`flask.Flask` application that uses
:class:`arxiv.base.Base`. Preference is given to each of those respective
sources with increasing priority.

This will build URLs with the ``https`` scheme by default. To use ``http``,
set ``EXTERNAL_URL_SCHEME = 'http'`` in your configuration.


Danger! Memory leaks lurk here
------------------------------
Earlier versions of this module built Werkzeug routing machinery (Rules, Maps,
etc) on the fly. This led to serious memory leaks. As of v0.15.6,
:class:`.Base` uses :func:`register_external_urls` to set up external URL
handling, which registers a single :class:`.MapAdapter` on a Flask app. This
adapter is in turn used by :func:`external_url_handler` on demand.

See ARXIVNG-2085.

"""

import sys
from typing import Dict, Any, List
from urllib.parse import parse_qs
from werkzeug.urls import url_encode, url_parse, url_unparse
from werkzeug.routing import Map, Rule, BuildError, MapAdapter
from flask import current_app, g, Flask

from arxiv.base.exceptions import ConfigurationError
from arxiv.base.converter import ArXivConverter
from arxiv.base import logging
from arxiv.base import config as base_config

from .clickthrough import clickthrough_url
from .links import urlize, urlizer, url_for_doi

logger = logging.getLogger(__name__)


def build_adapter(app: Flask) -> MapAdapter:
    """Build a :class:`.MapAdapter` from configured URLs."""
    # Get the base URLs (configured in this package).
    configured_urls = {url[0]: url for url in base_config.URLS}
    # Privilege ARXIV_URLs set on the application config.
    current_urls = app.config.get('URLS', [])
    if current_urls:
        configured_urls.update({url[0]: url for url in current_urls})
    url_map = Map([
        Rule(pattern, endpoint=name, host=host, build_only=True)
        for name, pattern, host in configured_urls.values()
    ], converters={'arxiv': ArXivConverter}, host_matching=True)

    scheme = app.config.get('EXTERNAL_URL_SCHEME', 'https')
    base_host = app.config.get('BASE_SERVER', 'arxiv.org')
    adapter: MapAdapter = url_map.bind(base_host, url_scheme=scheme)
    return adapter


def external_url_handler(err: BuildError, endpoint: str, values: Dict) -> str:
    """
    Attempt to handle failed URL building with :func:`external_url_for`.

    This gets attached to a Flask application via the
    :func:`flask.Flask.url_build_error_handlers` hook.
    """
    values.pop('_external')
    try:
        url: str = current_app.external_url_adapter.build(endpoint,
                                                          values=values,
                                                          force_external=True)
    except BuildError:
        # Re-raise the original BuildError, in context of original traceback.
        exc_type, exc_value, tb = sys.exc_info()
        if exc_value is err:
            raise exc_type(exc_value).with_traceback(tb)    # type: ignore
        else:
            raise err
    return url


def canonical_url(id: str, version: int = 0) -> str:
    """
    Generate the canonical URL for an arXiv identifier.

    This can be done from just the ID because the category is only needed if it
    is in the ID. id can be just the id or idv or cat/id or cat/idv.
    """
    # TODO: This should be better.
    # There should probably be something like INTERNAL_URL_SCHEMA
    # Also, /abs should probably be specified somewhere else
    # like arxiv.base.canonical
    scheme = current_app.config.get('EXTERNAL_URL_SCHEME', 'https')
    host = current_app.config.get('MAIN_SERVER', 'arxiv.org')
    if version:
        return f'{scheme}://{host}/abs/{id}v{version}'
    return f'{scheme}://{host}/abs/{id}'


def register_external_urls(app: Flask) -> None:
    """Register :func:`external_url_handler` on a Flask app."""
    app.external_url_adapter = build_adapter(app)
    app.url_build_error_handlers.append(external_url_handler)
